#include <algorithm>  // std::sort
#include <numeric>    // std::accumulate

#include "cdo_remap.h"

const auto is_not_equal = [](auto a, auto b) noexcept { return (a < b || b < a); };
const auto is_equal = [](auto a, auto b) noexcept { return !(a < b || b < a); };

extern "C"
{
#include "clipping.h"
#include "area.h"
#include "geometry.h"
#include "yac_interp_method_conserv.h"
}

struct CellSearch
{
  enum yac_edge_type *edgeType = nullptr;
  size_t numCellCorners = 0;
  size_t maxCells = 0;
  Varray<double> partialAreas;
  Varray<struct grid_cell> gridCells;
  Varray<struct grid_cell> overlapCells;
};

static void
cellsearch_realloc(size_t numCells, CellSearch &search)
{
  if (numCells > search.maxCells)
    {
      search.partialAreas.resize(numCells);
      search.overlapCells.resize(numCells);
      search.gridCells.resize(numCells);

      for (size_t i = search.maxCells; i < numCells; ++i)
        {
          search.overlapCells[i].array_size = 0;
          search.overlapCells[i].num_corners = 0;
          search.overlapCells[i].edge_type = nullptr;
          search.overlapCells[i].coordinates_xyz = nullptr;

          search.gridCells[i].array_size = search.numCellCorners;
          search.gridCells[i].num_corners = search.numCellCorners;
          search.gridCells[i].edge_type = search.edgeType;
          search.gridCells[i].coordinates_xyz = new double[search.numCellCorners][3];
        }

      search.maxCells = numCells;
    }
}

static void
cellsearch_free(CellSearch &search)
{
  for (size_t i = 0; i < search.maxCells; i++)
    {
      if (search.overlapCells[i].array_size > 0)
        {
          if (search.overlapCells[i].coordinates_xyz) free(search.overlapCells[i].coordinates_xyz);
          if (search.overlapCells[i].edge_type) free(search.overlapCells[i].edge_type);
        }

      delete[] search.gridCells[i].coordinates_xyz;
    }
}

static enum yac_cell_type
get_cell_type(struct grid_cell target_cell)
{
  /*
  YAC_ASSERT(
    target_cell.num_corners > 0, "ERROR(get_cell_type): CELL without edge")
  */
  enum yac_cell_type cell_type = MIXED_CELL;

  // if the cell is a typical lon-lat cell
  if ((target_cell.num_corners == 4)
      && ((target_cell.edge_type[0] == LAT_CIRCLE_EDGE && target_cell.edge_type[1] == LON_CIRCLE_EDGE
           && target_cell.edge_type[2] == LAT_CIRCLE_EDGE && target_cell.edge_type[3] == LON_CIRCLE_EDGE)
          || (target_cell.edge_type[0] == LON_CIRCLE_EDGE && target_cell.edge_type[1] == LAT_CIRCLE_EDGE
              && target_cell.edge_type[2] == LON_CIRCLE_EDGE && target_cell.edge_type[3] == LAT_CIRCLE_EDGE)))
    {

      cell_type = LON_LAT_CELL;
    }
  else
    {

      size_t count_lat_edges = 0, count_great_circle_edges = 0;

      // count the number of edges for each type
      // (lon edges are counted as gc ones)
      for (size_t i = 0; i < target_cell.num_corners; ++i)
        if (target_cell.edge_type[i] == LON_CIRCLE_EDGE || target_cell.edge_type[i] == GREAT_CIRCLE_EDGE)
          count_great_circle_edges++;
        else
          count_lat_edges++;

      // if the cell is a lon lat cell with one lat edge having a length of zero
      // due to being at a pole
      if ((count_lat_edges == 1) && (count_great_circle_edges == 2))
        {

          size_t i;
          for (i = 0; i < 3; ++i)
            if (target_cell.edge_type[i] == LAT_CIRCLE_EDGE) break;
          size_t pol_index = (3 + i - 1) % 3;
          if (fabs(fabs(target_cell.coordinates_xyz[pol_index][2]) - 1.0) < yac_angle_low_tol) cell_type = LON_LAT_CELL;

          // if the cell only consists of great circle edges
        }
      else if (count_lat_edges == 0)
        cell_type = GREAT_CIRCLE_CELL;

      // if the cell only consists of lat circle edges
      else if (count_great_circle_edges == 0)
        cell_type = LAT_CELL;
    }
  /*
  YAC_ASSERT(
    cell_type != MIXED_CELL,
    "invalid cell type (cell contains edges consisting "
    "of great circles and circles of latitude)")
  */
  return cell_type;
}

static inline double
gridcell_area(const grid_cell &cell)
{
  return yac_huiliers_area(cell);
}

static void
cdo_compute_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell)
{
  auto &overlapCells = search.overlapCells;

  // Do the clipping and get the cell for the overlapping area
  yac_cell_clipping(numSearchCells, search.gridCells.data(), gridCell.yacGridCell, overlapCells.data());

  // Get the partial areas for the overlapping regions
  for (size_t i = 0; i < numSearchCells; i++) search.partialAreas[i] = gridcell_area(overlapCells[i]);

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area : %lf\n", search.partialAreas[i]);
#endif
}
/*
static void
cdo_compute_concave_overlap_info(size_t numSearchCells, CellSearch &search, const GridCell &gridCell, double tgtLon, double tgtLat,
                                 double *overlapAreas, double (*overlapBarycenters)[3])
{
  auto &overlapCells = search.overlapCells;
  auto &sourceCells = search.gridCells;
  auto targetCell = gridCell.yacGridCell;

  if (overlapBarycenters != NULL)
    for (size_t i = 0; i < numSearchCells; ++i)
      for (int j = 0; j < 3; ++j) overlapBarycenters[i][j] = 0.0;

  auto target_cell_type = (targetCell.num_corners > 3) ? get_cell_type(targetCell) : MIXED_CELL;

  if (targetCell.num_corners < 4 || target_cell_type == LON_LAT_CELL)
    {
      yac_cell_clipping(numSearchCells, sourceCells.data(), targetCell, overlapCells.data());
      for (size_t i = 0; i < numSearchCells; ++i)
        {
          if (overlapCells[i].num_corners > 1)
            {
              if (overlapBarycenters == NULL)
                overlapAreas[i] = yac_huiliers_area(overlapCells[i]);
              else
                {
                  overlapAreas[i] = yac_huiliers_area_info(overlapCells[i], overlapBarycenters[i]);
                  normalise_vector(overlapBarycenters[i]);
                }
            }
          else { overlapAreas[i] = 0.0; }
        }
      return;
    }

  double coordinates_xyz[3][3] = { { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 } };
  enum yac_edge_type edgeTypes[3] = { GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE };

  struct grid_cell targetPartialCell;
  targetPartialCell.array_size = 3;
  targetPartialCell.num_corners = 3;
  targetPartialCell.coordinates_xyz = coordinates_xyz;
  targetPartialCell.edge_type = edgeTypes;

  // Do the clipping and get the cell for the overlapping area

  for (size_t i = 0; i < numSearchCells; i++) partialAreas[i] = 0.0;

  // common node point to all partial target cells
  auto partial_cell_xyz = targetPartialCell.coordinates_xyz;
  auto cell_xyz = targetCell.coordinates_xyz;

  LL_to_XYZ(tgtLon, tgtLat, partial_cell_xyz[0]);

  for (size_t cornerNum = 0; cornerNum < targetCell.num_corners; ++cornerNum)
    {
      auto cornerA = cornerNum;
      auto cornerB = (cornerNum + 1) % targetCell.num_corners;

      // skip clipping and area calculation for degenerated triangles
      //
      // If this is not sufficient, instead we can try something like:
      //
      //     point_list target_list
      //     init_point_list(&target_list);
      //     generate_point_list(&target_list, targetCell);
      //     grid_cell temp_target_cell;
      //     generate_overlap_cell(target_list, temp_target_cell);
      //     free_point_list(&target_list);
      //
      // and use temp_target_cell for triangulation.
      //
      // Compared to the if statement below the alternative seems to be quite costly.

      if (points_are_identically(cell_xyz[cornerA], cell_xyz[cornerB])) continue;

      for (int k = 0; k < 3; ++k) partial_cell_xyz[1][k] = cell_xyz[cornerA][k];
      for (int k = 0; k < 3; ++k) partial_cell_xyz[2][k] = cell_xyz[cornerB][k];

      yac_cell_clipping(numSearchCells, sourceCells.data(), targetPartialCell, overlapCells.data());

      // Get the partial areas for the overlapping regions as sum over the partial target cells.
      for (size_t i = 0; i < numSearchCells; i++)
        {
          if (overlapCells[i].num_corners == 0) continue;

          if (overlapBarycenters == NULL)
            overlapAreas[i] += gridcell_area(overlapCells[i]);
          else
            overlapAreas[i] += yac_huiliers_area_info(overlapCells[i], overlapBarycenters[i]);
        }
    }

  if (overlapBarycenters != NULL)
    for (size_t i = 0; i < numSearchCells; i++)
      if (overlapAreas[i] > 0.0) normalise_vector(overlapBarycenters[i]);

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area %zu: %lf\n", i, overlapAreas[i]);
#endif
}

static void
cdo_compute_concave_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell, double tgtLon, double tgtLat)
{
  auto partialAreas = search.partialAreas.data();
  cdo_compute_concave_overlap_info(numSearchCells, search, gridCell, tgtLon, tgtLat, partialAreas, NULL);
}
*/

static double
get_edge_direction(const double *ref_corner, double *corner_a, double *corner_b)
{
  double edge_norm[3];
  crossproduct_kahan(corner_a, corner_b, edge_norm);
  normalise_vector(edge_norm);

  // sine of the angle between the edge and the reference corner
  double angle = edge_norm[0] * ref_corner[0] + edge_norm[1] * ref_corner[1] + edge_norm[2] * ref_corner[2];

  // if the reference corner is directly on the edge
  // (for small angles sin(x)==x)
  if (fabs(angle) < yac_angle_tol) return 0.0;

  return copysign(1.0, angle);
}

static void
cdo_compute_concave_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell)
{
  auto targetCell = gridCell.yacGridCell;
  auto &overlapAreas = search.partialAreas;
  auto &overlapCells = search.overlapCells;
  auto &sourceCells = search.gridCells;

  // common node point to all partial target cells
  double *baseCorner = targetCell.coordinates_xyz[0];

  // the triangulation algorithm only works for cells that only have great circle edges
  enum yac_edge_type edgeTypes[3] = { GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE };
  double coordinates_xyz[3][3] = { { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 } };
  for (int k = 0; k < 3; ++k) coordinates_xyz[0][k] = baseCorner[k];

  // data structure to hold the triangles of the target cell
  struct grid_cell partialCell;
  partialCell.array_size = 3;
  partialCell.num_corners = 3;
  partialCell.coordinates_xyz = coordinates_xyz;
  partialCell.edge_type = edgeTypes;

  // Do the clipping and get the cell for the overlapping area

  for (size_t i = 0; i < numSearchCells; ++i) overlapAreas[i] = 0.0;

  // for all triangles of the target cell
  // (triangles a formed by first corner of the target cells and each edge of
  //  the cell; the first and last edge of the cell already, contain the
  //  first corner, therefore we can skip them)
  for (size_t cornerIdx = 1; cornerIdx < targetCell.num_corners - 1; ++cornerIdx)
    {
      auto cornerA = targetCell.coordinates_xyz[cornerIdx];
      auto cornerB = targetCell.coordinates_xyz[(cornerIdx + 1)];

      // if the current edge has a length of zero
      if (points_are_identically(cornerA, cornerB)) continue;

      auto edgeDirection = get_edge_direction(baseCorner, cornerA, cornerB);

      for (int k = 0; k < 3; ++k) partialCell.coordinates_xyz[1][k] = cornerA[k];
      for (int k = 0; k < 3; ++k) partialCell.coordinates_xyz[2][k] = cornerB[k];

      // clip the current target cell triangle with all source cells
      yac_cell_clipping(numSearchCells, sourceCells.data(), partialCell, overlapCells.data());

      // Get the partial areas for the overlapping regions as sum over the partial target cells.
      for (size_t i = 0; i < numSearchCells; ++i)
        {
          if (overlapCells[i].num_corners == 0) continue;

          overlapAreas[i] += gridcell_area(overlapCells[i]) * edgeDirection;
        }
    }

  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (overlapAreas[i] < 0.0) overlapAreas[i] = -overlapAreas[i];
    }

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area %zu: %lf\n", i, overlapAreas[i]);
#endif
}

static void
set_cell_coordinates_yac(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo,
                         const struct grid_cell &yacGridCell, double *x, double *y)
{
  auto storeXY = (x && y);
  auto xyz = yacGridCell.coordinates_xyz;

  const auto cellCornersLon = &gridInfo.cellCornersLon[cellIndex * numCorners];
  const auto cellCornersLat = &gridInfo.cellCornersLat[cellIndex * numCorners];
  for (size_t i = 0; i < numCorners; ++i) LL_to_XYZ(cellCornersLon[i], cellCornersLat[i], xyz[i]);

  if (storeXY)
    {
      for (size_t k = 0; k < numCorners; ++k)
        {
          x[k] = cellCornersLon[k];
          y[k] = cellCornersLat[k];
        }
    }
}

static void
set_cell_coordinates(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo, const GridCell &gridCell)
{
  set_cell_coordinates_yac(cellIndex, numCorners, gridInfo, gridCell.yacGridCell, gridCell.coordinates_x, gridCell.coordinates_y);
}

static void
set_coordinates_yac(const size_t numCells, const Varray<size_t> &cellIndices, const size_t numCorners, const GridInfo &gridInfo,
                    const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    set_cell_coordinates_yac(cellIndices[i], numCorners, gridInfo, yacGridCell[i], nullptr, nullptr);
}

static void
set_coordinates_yac(const double (*xyzCoords)[3], const size_t numCells, const Varray<size_t> &cellIndices, const size_t numCorners,
                    const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    {
      auto offset = cellIndices[i] * numCorners;
      auto xyz = yacGridCell[i].coordinates_xyz;
      for (size_t k = 0; k < numCorners; ++k)
        for (size_t l = 0; l < 3; ++l) xyz[k][l] = xyzCoords[offset + k][l];
    }
}

static void
gridcell_init_yac(GridCell &gridCell, size_t numCorners, enum yac_edge_type *edgeType)
{
  gridCell.yacGridCell.array_size = numCorners;
  gridCell.yacGridCell.num_corners = numCorners;
  gridCell.yacGridCell.edge_type = edgeType;
  gridCell.yacGridCell.coordinates_xyz = new double[numCorners][3];
  gridCell.coordinates_x = new double[numCorners];
  gridCell.coordinates_y = new double[numCorners];
}

static void
gridcell_free_yac(const GridCell &gridCell)
{
  delete[] gridCell.yacGridCell.coordinates_xyz;
  delete[] gridCell.coordinates_x;
  delete[] gridCell.coordinates_y;
}

static size_t
remove_invalid_areas(size_t numSearchCells, Varray<double> &areas, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (areas[i] > 0.0)
        {
          areas[n] = areas[i];
          searchIndices[n] = searchIndices[i];
          n++;
        }
    }

  return n;
}

static size_t
remove_invalid_weights(size_t gridSize, size_t numWeights, Varray<double> &weights, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      auto cellIndex = (weights[i] > 0.0) ? searchIndices[i] : gridSize;
      if (cellIndex != gridSize)
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static size_t
remove_unmask_weights(const Varray<short> &gridMask, size_t numWeights, Varray<double> &weights, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      auto cellIndex = searchIndices[i];
      /*
        Store the appropriate addresses and weights.
        Also add contributions to cell areas.
        The source grid mask is the master mask.
      */
      if (gridMask[cellIndex])
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static void
normalize_weights(NormOpt normOpt, double cellArea, double cellFrac, size_t numWeights, Varray<double> &weights)
{
  if (normOpt == NormOpt::DESTAREA)
    {
      auto normFactor = is_not_equal(cellArea, 0.0) ? 1.0 / cellArea : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::FRACAREA)
    {
      auto normFactor = is_not_equal(cellFrac, 0.0) ? 1.0 / cellFrac : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::NONE) {}
}

static void
correct_weights(double cellArea, size_t numWeights, Varray<double> &weights)
{
  for (size_t i = 0; i < numWeights; ++i) weights[i] /= cellArea;
  yac_correct_weights(numWeights, weights.data());
  for (size_t i = 0; i < numWeights; ++i) weights[i] *= cellArea;
}

static void
sort_weights(size_t numWeights, Varray<size_t> &searchIndices, Varray<double> &weights)
{
  size_t n;
  for (n = 1; n < numWeights; ++n)
    if (searchIndices[n] < searchIndices[n - 1]) break;
  if (n == numWeights) return;

  if (numWeights > 1)
    {
      struct IndexWeightX
      {
        size_t index;
        double weight;
      };

      Varray<IndexWeightX> indexWeights(numWeights);

      for (n = 0; n < numWeights; ++n)
        {
          indexWeights[n].index = searchIndices[n];
          indexWeights[n].weight = weights[n];
        }

      const auto compareIndex = [](const auto &a, const auto &b) noexcept { return a.index < b.index; };
      std::sort(indexWeights.begin(), indexWeights.end(), compareIndex);

      for (n = 0; n < numWeights; ++n)
        {
          searchIndices[n] = indexWeights[n].index;
          weights[n] = indexWeights[n].weight;
        }
    }
}

static int
get_lonlat_circle_index(size_t numCells, size_t numCorners, const Varray<double> &clon, const Varray<double> &clat)
{
  int lonlatCircleIndex = -1;

  if (numCorners == 4)
    {
      size_t iadd = (numCells < 100) ? 1 : numCells / 30 - 1;
      size_t num_i = 0, num_eq0 = 0, num_eq1 = 0;

      for (size_t i = 0; i < numCells; i += iadd)
        {
          auto i4 = i * 4;
          num_i++;
          // clang-format off
          if (is_equal(clon[i4 + 1], clon[i4 + 2]) && is_equal(clon[i4 + 3], clon[i4 + 0]) &&
              is_equal(clat[i4 + 0], clat[i4 + 1]) && is_equal(clat[i4 + 2], clat[i4 + 3]))
            {
              num_eq1++;
            }
          else if (is_equal(clon[i4 + 0], clon[i4 + 1]) && is_equal(clon[i4 + 2], clon[i4 + 3]) &&
                   is_equal(clat[i4 + 1], clat[i4 + 2]) && is_equal(clat[i4 + 3], clat[i4 + 0]))
            {
              num_eq0++;
            }
          // clang-format on
        }

      if (num_i == num_eq1) lonlatCircleIndex = 1;
      if (num_i == num_eq0) lonlatCircleIndex = 0;
    }

  // printf("lonlatCircleIndex %d\n", lonlatCircleIndex);

  return lonlatCircleIndex;
}

static int
get_lonlat_circle_index(const GridInfo &gridInfo)
{
  return get_lonlat_circle_index(gridInfo.numCells, gridInfo.numCorners, gridInfo.cellCornersLon, gridInfo.cellCornersLat);
}

static double
conserv_remap(const Varray<double> &srcArray, size_t numWeights, const Varray<double> &weights, const Varray<size_t> &srcIndices)
{
  double tgtPoint = 0.0;
  for (size_t i = 0; i < numWeights; ++i) tgtPoint += srcArray[srcIndices[i]] * weights[i];

  return tgtPoint;
}

void
remap_conserv2(NormOpt normOpt, const RemapSearch &remapSearch, const Varray<double> &srcArray, Varray<double> &tgtArray,
               double missval)
{
  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  auto srcNumCells = srcGrid.numCells;
  auto tgtNumCells = tgtGrid.numCells;

  Varray<short> srcGridMask(srcNumCells);
  for (size_t i = 0; i < srcNumCells; ++i) srcGridMask[i] = !is_equal(srcArray[i], missval);

  auto srcNumCorners = srcGrid.numCorners;
  auto tgtNumCorners = tgtGrid.numCorners;

  enum yac_edge_type lonlatCircleType[] = { LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE };
  Varray<enum yac_edge_type> greatCircleType(std::max(srcNumCorners, tgtNumCorners), GREAT_CIRCLE_EDGE);

  auto srcEdgeType = greatCircleType.data();
  auto tgtEdgeType = greatCircleType.data();

  enum yac_cell_type tgtCellType = MIXED_CELL;

  if (srcNumCorners == 4)
    {
      auto lonlatCircleIndex = get_lonlat_circle_index(srcGrid);
      if (lonlatCircleIndex >= 0) srcEdgeType = &lonlatCircleType[lonlatCircleIndex];
    }

  if (tgtNumCorners == 4)
    {
      auto lonlatCircleIndex = get_lonlat_circle_index(tgtGrid);
      if (lonlatCircleIndex >= 0)
        {
          tgtCellType = LON_LAT_CELL;
          tgtEdgeType = &lonlatCircleType[lonlatCircleIndex];
        }
    }

  GridCell tgtGridCell;
  gridcell_init_yac(tgtGridCell, tgtNumCorners, tgtEdgeType);

  CellSearch cellSearch;
  cellSearch.numCellCorners = srcNumCorners;
  cellSearch.edgeType = srcEdgeType;

  auto numCellCorners = srcNumCorners;  // num of corners of search cells

  Varray<size_t> searchIndices(srcNumCells);

  // Loop over target grid
  for (size_t tgtCellIndex = 0; tgtCellIndex < tgtNumCells; ++tgtCellIndex)
    {
      tgtArray[tgtCellIndex] = missval;

      // if (!tgtGrid.mask[tgtCellIndex]) continue;

      set_cell_coordinates(tgtCellIndex, tgtNumCorners, tgtGrid, tgtGridCell);

      // Get search cells
      auto numSearchCells = do_gridcell_search(remapSearch.gcs, false, tgtGridCell, searchIndices);

      if (numSearchCells == 0) continue;

      // Create search arrays

      cellsearch_realloc(numSearchCells, cellSearch);

      // set_coordinates_yac(numSearchCells, searchIndices, numCellCorners, srcGrid, cellSearch.gridCells);
      if (remapSearch.gcs.xyzCoords)
        set_coordinates_yac(remapSearch.gcs.xyzCoords, numSearchCells, searchIndices, numCellCorners, cellSearch.gridCells);
      else
        set_coordinates_yac(numSearchCells, searchIndices, numCellCorners, srcGrid, cellSearch.gridCells);

      if (tgtNumCorners < 4 || tgtCellType == LON_LAT_CELL)
        cdo_compute_overlap_areas(numSearchCells, cellSearch, tgtGridCell);
      else
        cdo_compute_concave_overlap_areas(numSearchCells, cellSearch, tgtGridCell);

      // For all supermesh cells compute the area and normalised area.
      // Additionally, remove all empty supermesh cells.
      size_t new_num_super_cells = 0;
      size_t tgt_idx = 0;
      /*
      for (size_t i = 0, j = 0; i < total_num_overlaps;)
        {

          // get information about the current target cell
          size_t curr_tgt_cell = super_cells[i].tgt.local_id;
          yac_const_basic_grid_data_get_grid_cell(tgt_basic_grid_data, curr_tgt_cell, &tgt_grid_cell);

          double curr_tgt_cell_coverage = 0.0;

          // for all supermesh cells overlapping with the current target cell
          for (;(i < total_num_overlaps) && (super_cells[i].tgt.local_id == curr_tgt_cell); ++i)
            {

              // get the current source cell
              yac_const_basic_grid_data_get_grid_cell(src_basic_grid_data, super_cells[i].src.local_id, &src_grid_cell);

              // compute area of the current supermesh cell
              double super_cell_area;
              double barycenter[3];
              cdo_compute_concave_overlap_info(1, cellSearch, tgtGridCell,
                                               tgtGrid.cellCentersLon[tgtCellIndex], tgtGrid.cellCentersLat[tgtCellIndex],
                                               &super_cell_area, &barycenter);
            }

          // if there is an overlap between the current source and target cell
          if (super_cell_area > 0.0)
            {
              super_cells[new_num_super_cells].src = super_cells[i].src;
              super_cells[new_num_super_cells].tgt = super_cells[i].tgt;
              super_cells[new_num_super_cells].area = super_cell_area;
              memcpy(super_cells[new_num_super_cells].barycenter, barycenter, 3 * sizeof(double));
              super_cells[new_num_super_cells].src_cell_gradient = NULL;
              ++new_num_super_cells;

              curr_tgt_cell_coverage += super_cell_area;
            }
        }
      */
      auto &partialWeights = cellSearch.partialAreas;

      auto numWeights = remove_invalid_areas(numSearchCells, partialWeights, searchIndices);

      auto tgtCellArea = gridcell_area(tgtGridCell.yacGridCell);

      if (normOpt == NormOpt::FRACAREA) correct_weights(tgtCellArea, numWeights, partialWeights);

      numWeights = remove_invalid_weights(srcNumCells, numWeights, partialWeights, searchIndices);
      numWeights = remove_unmask_weights(srcGridMask, numWeights, partialWeights, searchIndices);

      auto tgtCellFrac = std::accumulate(partialWeights.begin(), partialWeights.begin() + numWeights, 0.0);

      if (numWeights)
        {
          sort_weights(numWeights, searchIndices, partialWeights);
          // Normalize weights using target cell area if requested
          normalize_weights(normOpt, tgtCellArea, tgtCellFrac, numWeights, partialWeights);
          tgtArray[tgtCellIndex] = conserv_remap(srcArray, numWeights, partialWeights, searchIndices);
        }
    }

  // Finished with all cells: deallocate search arrays

  cellsearch_free(cellSearch);
  gridcell_free_yac(tgtGridCell);
}  // remap_conserv2
